@@ -4,28 +4,30 @@ import docker
 import os
 import time
 import subprocess
-from docker.client import DockerClient
 
+from typing import Callable, Type, Iterable
 from threading import Thread
 from functools import wraps
 from colorama import (Fore, Style)
 from sqlalchemy.engine.url import make_url
+from docker.client import DockerClient
 
 from vantage6.common import (info, warning, error, debug as debug_msg,
-                             check_config_write_permissions)
-from vantage6.common.docker_addons import (
-    pull_if_newer, remove_container_if_exists,
-    get_server_config_name
+                             check_config_writeable)
+from vantage6.common.docker.addons import (
+    pull_if_newer, check_docker_running, remove_container,
+    get_server_config_name, get_container, get_num_nonempty_networks,
+    get_network, delete_network
 )
-from vantage6.common.network_manager import NetworkManager
+from vantage6.common.docker.network_manager import NetworkManager
 from vantage6.common.globals import (
     APPNAME,
     STRING_ENCODING,
     DEFAULT_DOCKER_REGISTRY,
     DEFAULT_SERVER_IMAGE
 )
+from vantage6.cli.rabbitmq import split_rabbitmq_uri
 
-# from vantage6.cli import fixture
 from vantage6.cli.globals import (DEFAULT_SERVER_ENVIRONMENT,
                                   DEFAULT_SERVER_SYSTEM_FOLDERS)
 from vantage6.cli.context import ServerContext
@@ -33,15 +35,28 @@ from vantage6.cli.configuration_wizard import (
     select_configuration_questionaire,
     configuration_wizard
 )
+from vantage6.cli.utils import check_config_name_allowed
 from vantage6.cli.rabbitmq.queue_manager import RabbitMQManager
-from vantage6.cli import __version__
+from vantage6.cli import __version__, rabbitmq
 
 
-def click_insert_context(func):
+def click_insert_context(func: Callable) -> Callable:
+    """
+    Supply the Click function with additional context parameters. The context
+    is then passed to the function as the first argument.
 
-    # add option decorators
+    Parameters
+    ----------
+    func : Callable
+        function you want the context to be passed to
+
+    Returns
+    -------
+    Callable
+        Click function with context
+    """
     @click.option('-n', '--name', default=None,
-                  help="name of the configutation you want to use.")
+                  help="name of the configuration you want to use.")
     @click.option('-c', '--config', default=None,
                   help='absolute path to configuration-file; overrides NAME')
     @click.option('-e', '--environment',
@@ -51,58 +66,57 @@ def click_insert_context(func):
     @click.option('--user', 'system_folders', flag_value=False,
                   default=DEFAULT_SERVER_SYSTEM_FOLDERS)
     @wraps(func)
-    def func_with_context(name, config, environment, system_folders,
-                          *args, **kwargs):
+    def func_with_context(name: str, config: str, environment: str,
+                          system_folders: bool, *args, **kwargs) -> Callable:
+        """
+        Decorator function that adds the context to the function.
 
-        # select configuration if none supplied
+        Parameters
+        ----------
+        name : str
+            name of the configuration you want to use.
+        config : str
+            path to configuration file, overrides name
+        environment : str
+            DTAP environment to use
+        system_folders : bool
+            Wether to use system folders or not
+
+        Returns
+        -------
+        Callable
+            Decorated function
+        """
+        # path to configuration file always overrides name
         if config:
             ctx = ServerContext.from_external_config_file(
                 config,
                 environment,
                 system_folders
             )
-        else:
-            if name:
-                name, environment = (name, environment)
-            else:
-                try:
-                    name, environment = select_configuration_questionaire(
-                        "server", system_folders
-                    )
-                except Exception:
-                    error("No configurations could be found!")
-                    exit(1)
+            return func(ctx, *args, **kwargs)
 
-            # raise error if config could not be found
-            if not ServerContext.config_exists(
-                name,
-                environment,
-                system_folders
-            ):
-                scope = "system" if system_folders else "user"
-                error(
-                    f"Configuration {Fore.RED}{name}{Style.RESET_ALL} with "
-                    f"{Fore.RED}{environment}{Style.RESET_ALL} does not exist "
-                    f"in the {Fore.RED}{scope}{Style.RESET_ALL} folders!"
+        # in case no name is supplied, ask the user to select one
+        if not name:
+            try:
+                # select configuration if none supplied
+                name, environment = select_configuration_questionaire(
+                    "server", system_folders
                 )
+            except Exception:
+                error("No configurations could be found!")
                 exit(1)
 
-            # create server context, and initialize db
-            ServerContext.LOGGING_ENABLED = False
-            ctx = ServerContext(
-                name,
-                environment=environment,
-                system_folders=system_folders
-            )
-
+        ctx = get_server_context(name, environment, system_folders)
         return func(ctx, *args, **kwargs)
+
     return func_with_context
 
 
 @click.group(name='server')
-def cli_server():
+def cli_server() -> None:
     """Subcommand `vserver`."""
-    pass
+
 
 #
 #   start
@@ -120,15 +134,42 @@ def cli_server():
 @click.option('--attach/--detach', default=False,
               help="Attach server logs to the console after start")
 @click_insert_context
-def cli_server_start(ctx, ip, port, image, rabbitmq_image, keep, mount_src,
-                     attach):
-    """Start the server."""
+def cli_server_start(ctx: Type[ServerContext], ip: str, port: int, image: str,
+                     rabbitmq_image: str, keep: bool, mount_src: str,
+                     attach: bool) -> None:
+    """
+    Start the server in a Docker container.
 
+    Parameters
+    ----------
+    ctx : Type[ServerContext]
+        Server context object
+    ip : str
+        ip interface to listen on
+    port : int
+        port to listen on
+    image : str
+        Server Docker image to use
+    rabbitmq_image : str
+        RabbitMQ docker image to use
+    keep : bool
+        Wether to keep the image after the server has finished, useful for
+        debugging
+    mount_src : str
+        Path to the vantage6 package source, this overrides the source code in
+        the container. This is useful when developing and testing the server.
+    attach : bool
+        Wether to attach the server logs to the console after starting the
+        server.
+    """
     info("Starting server...")
     info("Finding Docker daemon.")
     docker_client = docker.from_env()
     # will print an error if not
-    check_if_docker_deamon_is_running(docker_client)
+    check_docker_running()
+
+    # check if name is allowed for docker volume, else exit
+    check_config_name_allowed(ctx.name)
 
     # check that this server is not already running
     running_servers = docker_client.containers.list(
@@ -151,8 +192,9 @@ def cli_server_start(ctx, ip, port, image, rabbitmq_image, keep, mount_src,
     try:
         pull_if_newer(docker.from_env(), image)
         # docker_client.images.pull(image)
-    except Exception:
-        warning("... alas, no dice!")
+    except Exception as e:
+        warning(' ... Getting latest node image failed:')
+        warning(f"     {e}")
     else:
         info(" ... success!")
 
@@ -167,12 +209,11 @@ def cli_server_start(ctx, ip, port, image, rabbitmq_image, keep, mount_src,
     if mount_src:
         mount_src = os.path.abspath(mount_src)
         mounts.append(docker.types.Mount("/vantage6", mount_src, type="bind"))
-
     # FIXME: code duplication with cli_server_import()
     # try to mount database
     uri = ctx.config['uri']
     url = make_url(uri)
-    environment_vars = {"VANTAGE6_CONFIG_NAME": ctx.config_file_name}
+    environment_vars = None
 
     # If host is None, we're dealing with a file-based DB, like SQLite
     if (url.host is None):
@@ -191,10 +232,13 @@ def cli_server_start(ctx, ip, port, image, rabbitmq_image, keep, mount_src,
             "/mnt/database/", dirname, type="bind"
         ))
 
-        environment_vars["VANTAGE6_DB_URI"] = \
-            f"sqlite:////mnt/database/{basename}"
+        environment_vars = {
+            "VANTAGE6_DB_URI": f"sqlite:////mnt/database/{basename}",
+            "VANTAGE6_CONFIG_NAME": ctx.config_file_name
+        }
+
     else:
-        warning(f"Database could not be transfered, make sure {url.host} "
+        warning(f"Database could not be transferred, make sure {url.host} "
                 "is reachable from the Docker container")
         info("Consider using the docker-compose method to start a server")
 
@@ -240,7 +284,7 @@ def cli_server_start(ctx, ip, port, image, rabbitmq_image, keep, mount_src,
         network=server_network_mgr.network_name
     )
 
-    info(f"Success! container id = {container}")
+    info(f"Success! container id = {container.id}")
 
     if attach:
         logs = container.attach(stream=True, logs=True, stdout=True)
@@ -250,31 +294,21 @@ def cli_server_start(ctx, ip, port, image, rabbitmq_image, keep, mount_src,
                 time.sleep(1)
             except KeyboardInterrupt:
                 info("Closing log file. Keyboard Interrupt.")
+                info("Note that your server is still running! Shut it down "
+                     f"with {Fore.RED}vserver stop{Style.RESET_ALL}")
                 exit(0)
 
 
-def _start_rabbitmq(ctx: ServerContext, rabbitmq_image: str,
-                    network_mgr: NetworkManager) -> None:
-    """ Starts a RabbitMQ container """
-    if not ctx.config.get('rabbitmq'):
-        warning('Message queue disabled! This means that the server '
-                'application cannot scale horizontally!')
-    else:
-        # kick off RabbitMQ container
-        rabbit_mgr = RabbitMQManager(
-            ctx=ctx, network_mgr=network_mgr, image=rabbitmq_image)
-        rabbit_mgr.start()
-
-
 #
 #   list
 #
 @cli_server.command(name='list')
-def cli_server_configuration_list():
-    """Print the available configurations."""
-
+def cli_server_configuration_list() -> None:
+    """
+    Print the available configurations.
+    """
     client = docker.from_env()
-    check_if_docker_deamon_is_running(client)
+    check_docker_running()
 
     running_server = client.containers.list(
         filters={"label": f"{APPNAME}-type=server"})
@@ -323,37 +357,58 @@ def cli_server_configuration_list():
         warning(
              f"{Fore.RED}Failed imports: {len(f1)+len(f2)}{Style.RESET_ALL}")
 
+
 #
 #   files
 #
 @cli_server.command(name='files')
 @click_insert_context
-def cli_server_files(ctx):
-    """List files locations of a server instance."""
+def cli_server_files(ctx: Type[ServerContext]) -> None:
+    """
+    List files locations of a server instance.
+
+    Parameters
+    ----------
+    ctx : Type[ServerContext]
+        Server context object
+    """
     info(f"Configuration file = {ctx.config_file}")
     info(f"Log file           = {ctx.log_file}")
     info(f"Database           = {ctx.get_database_uri()}")
 
+
 #
 #   new
 #
 @cli_server.command(name='new')
 @click.option('-n', '--name', default=None,
-              help="name of the configutation you want to use.")
+              help="name of the configuration you want to use.")
 @click.option('-e', '--environment', default=DEFAULT_SERVER_ENVIRONMENT,
               help='configuration environment to use')
 @click.option('--system', 'system_folders', flag_value=True)
 @click.option('--user', 'system_folders', flag_value=False,
               default=DEFAULT_SERVER_SYSTEM_FOLDERS)
-def cli_server_new(name, environment, system_folders):
-    """Create new configuration."""
+def cli_server_new(name: str, environment: str, system_folders: bool) -> None:
+    """
+    Create a new server configuration.
 
+    Parameters
+    ----------
+    name : str
+        name of the new configuration
+    environment : str
+        DTAP environment you want to configure
+    system_folders : bool
+        Wether to use system folders or not
+    """
     if not name:
         name = q.text("Please enter a configuration-name:").ask()
-        name_new = name.replace(" ", "-")
-        if name != name_new:
+        if name.count(" ") > 0:
+            name = name.replace(" ", "-")
             info(f"Replaced spaces from configuration name: {name}")
-            name = name_new
+
+    # check if name is allowed for docker volume, else exit
+    check_config_name_allowed(name)
 
     # check that this config does not exist
     try:
@@ -365,31 +420,26 @@ def cli_server_new(name, environment, system_folders):
             )
             exit(1)
     except Exception as e:
-        print(e)
+        error(e)
         exit(1)
 
-     # Check that we can write in this folder
-    if not check_config_write_permissions(system_folders):
+    # Check that we can write in this folder
+    if not check_config_writeable(system_folders):
         error("Your user does not have write access to all folders. Exiting")
         info(f"Create a new server using '{Fore.GREEN}vserver new "
              "--user{Style.RESET_ALL}' instead!")
         exit(1)
 
     # create config in ctx location
-    cfg_file = configuration_wizard(
-        "server",
-        name,
-        environment=environment,
-        system_folders=system_folders
-    )
+    cfg_file = configuration_wizard("server", name, environment=environment,
+                                    system_folders=system_folders)
     info(f"New configuration created: {Fore.GREEN}{cfg_file}{Style.RESET_ALL}")
 
     # info(f"root user created.")
     flag = "" if system_folders else "--user"
-    info(
-        f"You can start the server by running "
-        f"{Fore.GREEN}vserver start {flag}{Style.RESET_ALL}"
-    )
+    info(f"You can start the server by running {Fore.GREEN}vserver start "
+         f"{flag}{Style.RESET_ALL}")
+
 
 #
 #   import
@@ -399,21 +449,43 @@ def cli_server_new(name, environment, system_folders):
 @click.argument('file_', type=click.Path(exists=True))
 @click.option('--drop-all', is_flag=True, default=False)
 @click.option('-i', '--image', default=None, help="Node Docker image to use")
+@click.option('--mount-src', default='',
+              help="mount vantage6-master package source")
 @click.option('--keep/--auto-remove', default=False,
               help="Keep image after finishing")
 @click_insert_context
-def cli_server_import(ctx, file_, drop_all, image, keep):
-    """ Import organizations/collaborations/users and tasks.
+def cli_server_import(ctx: Type[ServerContext], file_: str, drop_all: bool,
+                      image: str, mount_src: str, keep: bool) -> None:
+    """
+    Batch import organizations/collaborations/users and tasks.
 
-        Especially usefull for testing purposes.
+    Parameters
+    ----------
+    ctx : Type[ServerContext]
+        Server context object
+    file_ : str
+        Yaml file containing the vantage6 formatted data to import
+    drop_all : bool
+        Wether to drop all data before importing
+    image : str
+        Node Docker image to use which contains the import script
+    mount_src : str
+        Vantage6 source location, this will overwrite the source code in the
+        container. Useful for debugging/development.
+    keep : bool
+        Wether to keep the image after finishing/crashing. Useful for
+        debugging.
     """
     info("Starting server...")
     info("Finding Docker daemon.")
     docker_client = docker.from_env()
     # will print an error if not
-    check_if_docker_deamon_is_running(docker_client)
+    check_docker_running()
 
-    # pull lastest Docker image
+    # check if name is allowed for docker volume, else exit
+    check_config_name_allowed(ctx.name)
+
+    # pull latest Docker image
     if image is None:
         image = ctx.config.get(
             "image",
@@ -422,8 +494,9 @@ def cli_server_import(ctx, file_, drop_all, image, keep):
     info(f"Pulling latest server image '{image}'.")
     try:
         docker_client.images.pull(image)
-    except Exception:
-        warning("... alas, no dice!")
+    except Exception as e:
+        warning(' ... Getting latest node image failed:')
+        warning(f"     {e}")
     else:
         info(" ... success!")
 
@@ -443,6 +516,10 @@ def cli_server_import(ctx, file_, drop_all, image, keep):
     url = make_url(uri)
     environment_vars = None
 
+    if mount_src:
+        mount_src = os.path.abspath(mount_src)
+        mounts.append(docker.types.Mount("/vantage6", mount_src, type="bind"))
+
     # If host is None, we're dealing with a file-based DB, like SQLite
     if (url.host is None):
         db_path = url.database
@@ -465,7 +542,7 @@ def cli_server_import(ctx, file_, drop_all, image, keep):
         }
 
     else:
-        warning(f"Database could not be transfered, make sure {url.host} "
+        warning(f"Database could not be transferred, make sure {url.host} "
                 "is reachable from the Docker container")
         info("Consider using the docker-compose method to start a server")
 
@@ -494,36 +571,30 @@ def cli_server_import(ctx, file_, drop_all, image, keep):
 
     info(f"Success! container id = {container.id}")
 
-    # print_log_worker(container.logs(stream=True))
-    # for log in container.logs(stream=True):
-    #     print(log.decode("utf-8"))
-    # info(f"Check logs files using {Fore.GREEN}docker logs {container.id}"
-    #      f"{Style.RESET_ALL}")
-
-    # info("Reading yaml file.")
-    # with open(file_) as f:
-    #     entities = yaml.safe_load(f.read())
-
-    # info("Adding entities to database.")
-    # fixture.load(entities, drop_all=drop_all)
-
 
 #
 #   shell
 #
 @cli_server.command(name='shell')
 @click_insert_context
-def cli_server_shell(ctx):
-    """ Run a iPython shell. """
+def cli_server_shell(ctx: Type[ServerContext]) -> None:
+    """
+    Run an iPython shell in the server container and attach the ORM. This
+    can be used to modify the database.
 
+    Parameters
+    ----------
+    ctx : Type[ServerContext]
+        Server context object
+    """
     docker_client = docker.from_env()
     # will print an error if not
-    check_if_docker_deamon_is_running(docker_client)
+    check_docker_running()
 
     running_servers = docker_client.containers.list(
         filters={"label": f"{APPNAME}-type=server"})
 
-    if not ctx.docker_container_name in [s.name for s in running_servers]:
+    if ctx.docker_container_name not in [s.name for s in running_servers]:
         error(f"Server {Fore.RED}{ctx.name}{Style.RESET_ALL} is not running?")
         return
 
@@ -539,16 +610,31 @@ def cli_server_shell(ctx):
 #   stop
 #
 @cli_server.command(name='stop')
-@click.option("-n", "--name", default=None, help="configuration name")
+@click.option("-n", "--name", default=None, help="Configuration name")
+@click.option('-e', '--environment', default=DEFAULT_SERVER_ENVIRONMENT,
+              help='configuration environment to use')
 @click.option('--system', 'system_folders', flag_value=True)
 @click.option('--user', 'system_folders', flag_value=False,
               default=DEFAULT_SERVER_SYSTEM_FOLDERS)
-@click.option('--all', 'all_servers', flag_value=True)
-def cli_server_stop(name, system_folders, all_servers):
-    """Stop a or all running server. """
+@click.option('--all', 'all_servers', flag_value=True, help="Stop all servers")
+def cli_server_stop(name: str, environment: str, system_folders: bool,
+                    all_servers: bool):
+    """
+    Stop one or all running server(s).
 
+    Parameters
+    ----------
+    name : str
+        Name of the server to stop
+    environment : str
+        DTAP environment to use
+    system_folders : bool
+        Wether to use system folders or not
+    all_servers : bool
+        Wether to stop all servers or not
+    """
     client = docker.from_env()
-    check_if_docker_deamon_is_running(client)
+    check_docker_running()
 
     running_servers = client.containers.list(
         filters={"label": f"{APPNAME}-type=server"})
@@ -561,43 +647,24 @@ def cli_server_stop(name, system_folders, all_servers):
 
     if all_servers:
         for container_name in running_server_names:
-            _stop_server_containers(client, container_name, system_folders)
+            _stop_server_containers(client, container_name, environment,
+                                    system_folders)
+        return
+
+    # make sure we have a configuration name to work with
+    if not name:
+        container_name = q.select("Select the server you wish to stop:",
+                                  choices=running_server_names).ask()
     else:
-        if not name:
-            container_name = q.select("Select the server you wish to stop:",
-                                      choices=running_server_names).ask()
-        else:
+        post_fix = "system" if system_folders else "user"
+        container_name = f"{APPNAME}-{name}-{post_fix}-server"
 
-            post_fix = "system" if system_folders else "user"
-            container_name = f"{APPNAME}-{name}-{post_fix}-server"
+    if container_name not in running_server_names:
+        error(f"{Fore.RED}{name}{Style.RESET_ALL} is not running!")
+        return
 
-        if container_name in running_server_names:
-            _stop_server_containers(client, container_name, system_folders)
-        else:
-            error(f"{Fore.RED}{name}{Style.RESET_ALL} is not running?")
-
-
-def _stop_server_containers(client: DockerClient, container_name: str,
-                            system_folders: bool) -> None:
-    """
-    Given a server's name, kill its docker container and related (RabbitMQ)
-    containers.
-    """
-    # kill the server
-    container = client.containers.get(container_name)
-    container.kill()
-    info(f"Stopped the {Fore.GREEN}{container_name}{Style.RESET_ALL} server.")
-
-    # find the configuration name from the docker container name
-    # server name is formatted as f"{APPNAME}-{self.name}-{self.scope}-server"
-    scope = "system" if system_folders else "user"
-    config_name = get_server_config_name(container_name, scope)
-
-    # kill the RabbitMQ container (if it exists)
-    rabbit_container_name = f'{APPNAME}-{config_name}-rabbitmq'
-    remove_container_if_exists(client, name=rabbit_container_name)
-    info(f"Stopped the {Fore.GREEN}{rabbit_container_name}{Style.RESET_ALL} "
-         "container.")
+    _stop_server_containers(client, container_name, environment,
+                            system_folders)
 
 
 #
@@ -608,11 +675,19 @@ def _stop_server_containers(client: DockerClient, container_name: str,
 @click.option('--system', 'system_folders', flag_value=True)
 @click.option('--user', 'system_folders', flag_value=False,
               default=DEFAULT_SERVER_SYSTEM_FOLDERS)
-def cli_server_attach(name, system_folders):
-    """Attach the logs from the docker container to the terminal."""
+def cli_server_attach(name: str, system_folders: bool) -> None:
+    """
+    Attach the logs from the docker container to the terminal.
 
+    Parameters
+    ----------
+    name : str
+        Configuration name
+    system_folders : bool
+        Wether to use system folders or not
+    """
     client = docker.from_env()
-    check_if_docker_deamon_is_running(client)
+    check_docker_running()
 
     running_servers = client.containers.list(
         filters={"label": f"{APPNAME}-type=server"})
@@ -634,38 +709,44 @@ def cli_server_attach(name, system_folders):
                 time.sleep(1)
             except KeyboardInterrupt:
                 info("Closing log file. Keyboard Interrupt.")
+                info("Note that your server is still running! Shut it down "
+                     f"with {Fore.RED}vserver stop{Style.RESET_ALL}")
                 exit(0)
     else:
         error(f"{Fore.RED}{name}{Style.RESET_ALL} was not running!?")
 
 
-def check_if_docker_deamon_is_running(docker_client):
-    try:
-        docker_client.ping()
-    except Exception:
-        error("Docker socket can not be found. Make sure Docker is running.")
-        exit()
-
-
 #
 #   version
 #
 @cli_server.command(name='version')
 @click.option("-n", "--name", default=None, help="configuration name")
 @click.option('--system', 'system_folders', flag_value=True)
-@click.option('--user', 'system_folders', flag_value=False, default=
-              DEFAULT_SERVER_SYSTEM_FOLDERS)
-def cli_server_version(name, system_folders):
-    """Returns current version of vantage6 services installed."""
+@click.option('--user', 'system_folders', flag_value=False,
+              default=DEFAULT_SERVER_SYSTEM_FOLDERS)
+def cli_server_version(name: str, system_folders: bool):
+    """
+    Print the version of the vantage6 services.
 
+    Parameters
+    ----------
+    name : str
+        Name of the server to inspect
+    system_folders : bool
+        Wether to use system folders or not
+    """
     client = docker.from_env()
-    check_if_docker_deamon_is_running(client)
+    check_docker_running()
 
     running_servers = client.containers.list(
-        filters={"label": f"{APPNAME}-type=node"})
-    running_server_names = [node.name for node in running_servers]
+        filters={"label": f"{APPNAME}-type=server"})
+    running_server_names = [server.name for server in running_servers]
 
     if not name:
+        if not running_server_names:
+            error("No servers are running! You can only check the version for "
+                  "servers that are running")
+            exit(1)
         name = q.select("Select the server you wish to inspect:",
                         choices=running_server_names).ask()
     else:
@@ -674,9 +755,131 @@ def cli_server_version(name, system_folders):
 
     if name in running_server_names:
         container = client.containers.get(name)
-        version = container.exec_run(cmd='vserver-local version', stdout = True)
-        click.echo({"server": version.output.decode('utf-8'), "cli":__version__})
+        version = container.exec_run(cmd='vserver-local version',
+                                     stdout=True)
+        click.echo({"server": version.output.decode('utf-8'),
+                    "cli": __version__})
+    else:
+        error(f"Server {name} is not running! Cannot provide version...")
 
-def print_log_worker(logs_stream):
+
+#
+# helper functions
+#
+def get_server_context(name: str, environment: str, system_folders: bool) \
+        -> ServerContext:
+    """
+    Load the server context from the configuration file.
+
+    Returns
+    -------
+    ServerContext
+        Server context object
+    """
+    if not ServerContext.config_exists(name, environment, system_folders):
+        scope = "system" if system_folders else "user"
+        error(
+                f"Configuration {Fore.RED}{name}{Style.RESET_ALL} with "
+                f"{Fore.RED}{environment}{Style.RESET_ALL} does not exist "
+                f"in the {Fore.RED}{scope}{Style.RESET_ALL} folders!"
+            )
+        exit(1)
+
+    # We do not want to log this here, we do this in the container and not on
+    # the host. We only want CLI logging here.
+    ServerContext.LOGGING_ENABLED = False
+
+    # create server context, and initialize db
+    ctx = ServerContext(name, environment=environment,
+                        system_folders=system_folders)
+
+    return ctx
+
+
+def _start_rabbitmq(ctx: ServerContext, rabbitmq_image: str,
+                    network_mgr: NetworkManager) -> None:
+    """
+    Start the RabbitMQ container if it is not already running.
+
+    Parameters
+    ----------
+    ctx : ServerContext
+        Server context object
+    rabbitmq_image : str
+        RabbitMQ image to use
+    network_mgr : NetworkManager
+        Network manager object
+    """
+    rabbit_uri = ctx.config.get('rabbitmq_uri')
+    if not rabbit_uri:
+        warning('Message queue disabled! This means that the server '
+                'application cannot scale horizontally!')
+    elif rabbitmq.is_local_address(rabbit_uri):
+        # kick off RabbitMQ container
+        rabbit_mgr = RabbitMQManager(
+            ctx=ctx, network_mgr=network_mgr, image=rabbitmq_image)
+        rabbit_mgr.start()
+    else:
+        info("Detected that the RabbitMQ service is a external service. "
+             "Assuming this service is up and running.")
+
+
+def _stop_server_containers(client: DockerClient, container_name: str,
+                            environment: str, system_folders: bool) -> None:
+    """
+    Given a server's name, kill its docker container and related (RabbitMQ)
+    containers.
+
+    Parameters
+    ----------
+    client : DockerClient
+        Docker client
+    container_name : str
+        Name of the server to stop
+    environment : str
+        DTAP environment to use
+    system_folders : bool
+        Wether to use system folders or not
+    """
+    # kill the server
+    container = client.containers.get(container_name)
+    container.kill()
+    info(f"Stopped the {Fore.GREEN}{container_name}{Style.RESET_ALL} server.")
+
+    # find the configuration name from the docker container name
+    # server name is formatted as f"{APPNAME}-{self.name}-{self.scope}-server"
+    scope = "system" if system_folders else "user"
+    config_name = get_server_config_name(container_name, scope)
+
+    ctx = get_server_context(config_name, environment, system_folders)
+
+    # delete the server network
+    network_name = f"{APPNAME}-{ctx.name}-{ctx.scope}-network"
+    network = get_network(client, name=network_name)
+    delete_network(network, kill_containers=False)
+
+    # kill RabbitMQ if it exists and no other servers are using to it (i.e. it
+    # is not in other docker networks with other containers)
+    rabbit_uri = ctx.config.get('rabbitmq_uri')
+    if rabbit_uri:
+        rabbit_container_name = split_rabbitmq_uri(
+            rabbit_uri=rabbit_uri)['host']
+        rabbit_container = get_container(client, name=rabbit_container_name)
+        if rabbit_container and \
+                get_num_nonempty_networks(rabbit_container) == 0:
+            remove_container(rabbit_container, kill=True)
+            info(f"Stopped the {Fore.GREEN}{rabbit_container_name}"
+                 f"{Style.RESET_ALL} container.")
+
+
+def print_log_worker(logs_stream: Iterable[bytes]) -> None:
+    """
+    Print the logs from the docker container to the terminal.
+
+    Parameters
+    ----------
+    logs_stream : Iterable[bytes]
+        Output of the `container.attach(.)` method
+    """
     for log in logs_stream:
         print(log.decode(STRING_ENCODING), end="")
